import logging

from lastipy.lastfm.library.top_tracks import fetch_top_tracks 
from lastipy.lastfm.recommendations.similar_tracks import fetch_similar_tracks
from lastipy.lastfm.library import period
from lastipy.track import Track
from lastipy.lastfm.library.recent_tracks import fetch_recent_tracks
from lastipy.recommendations.rating_calculator import calculate_ratings


def fetch_recommendations(
          user,
          api_key,
          recommendation_period=period.OVERALL,
          max_similar_tracks_per_top_track=100,
          blacklisted_artists=[],
          prefer_unheard_artists=True):
    """Fetches recommendations for the given user by fetching their top tracks, then getting tracks similar
    to them, and filtering out the user's recent tracks and blacklisted artists"""

    logging.info("Fetching top recommendations for " + user)

    top_tracks = fetch_top_tracks(user=user, api_key=api_key, a_period=recommendation_period)

    top_tracks_to_recommendations = {}
    recommendations = []
    for top_track in top_tracks:
        try:
            recommendations_for_current_track = fetch_similar_tracks(api_key, top_track, max_similar_tracks_per_top_track)
            if recommendations_for_current_track:
                recommendations = recommendations + recommendations_for_current_track
                top_tracks_to_recommendations[top_track] = recommendations_for_current_track
        except Exception as e:
            logging.error(f"Error occurred fetching similar tracks: " + str(e))

    recommendations = calculate_ratings(user=user,
                                api_key=api_key,
                                prefer_unheard_artists=prefer_unheard_artists,
                                top_tracks_to_recommendations=top_tracks_to_recommendations)

    logging.debug(f"Before filtering, fetched " + str(len(recommendations))
                    + " recommendations: " + str(recommendations))

    recommendations = _filter_out_recent_tracks(user, api_key, recommendations)

    recommendations = _filter_out_blacklisted_artists(blacklisted_artists, recommendations)

    # Filter out duplicates
    # TODO maybe duplicates should mean a greater chance of getting that recommendation in the playlist?
    recommendations = list(set(recommendations))

    logging.info(f"Fetched " + str(len(recommendations)) + " recommendations: " + str(recommendations))

    return recommendations

def _filter_out_blacklisted_artists(blacklisted_artists, recommendations):
    logging.info("Filtering out blacklisted artists (" + str(blacklisted_artists) + ")...")
    recommendations = [recommendation for recommendation in recommendations
                        if not any(recommendation.artist.lower() == blacklisted_artist.lower()
                                    for blacklisted_artist in blacklisted_artists)]
    return recommendations

def _filter_out_recent_tracks(user, api_key, recommendations):
    recent_tracks = fetch_recent_tracks(user, api_key)
    logging.info("Filtering out recent tracks from recommendations...")
    recommendations = [recommendation for recommendation in recommendations
                        if not any(Track.are_equivalent(recommendation, recent_track)
                                    for recent_track in recent_tracks)]
    return recommendations



import unittest
from lastipy.lastfm.library.top_track import TopTrack
from lastipy.lastfm.recommendations.recommended_track import RecommendedTrack
from unittest.mock import patch
from lastipy.lastfm.recommendations.similar_tracks import fetch_similar_tracks


class SimilarTracksFetcherTest(unittest.TestCase):
    
    @patch('requests.get')
    def test_track_has_multiple_similar_tracks(self, mock_requests_get):
        playcount = 5
        track_to_check = TopTrack(track_name="Night Fever", artist="Bee Gees", playcount=playcount)
        expected_similar_tracks = [
            RecommendedTrack(track_name="Stayin' Alive", artist="Bee Gees", recommendation_rating=playcount),
            RecommendedTrack(track_name="You Should Be Dancing", artist="Bee Gees", recommendation_rating=playcount)
        ]

        json_tracks = []
        for track in expected_similar_tracks:
            json_track = {
                'name': track.track_name,
                'artist': {
                    'name': track.artist
                },
                'match': track.recommendation_rating
            }
            json_tracks.append(json_track)

        json_response = {
            'similartracks': {
                'track': json_tracks
            }
        }
        
        mock_requests_get.return_value.json.return_value = json_response

        fetched_tracks = fetch_similar_tracks('', track_to_check, 2)
        self.assertCountEqual(expected_similar_tracks, fetched_tracks)
        for track in fetched_tracks:
            self.assertEqual(playcount, track.recommendation_rating)


import logging


#TODO test
def add_to_playlist(spotify, username, playlist_name, tracks):
    """Adds the given tracks to the given user's given playlist. If the playlist does not exist, creates it first"""

    playlists = spotify.current_user_playlists()
    playlists = [playlist for playlist in playlists['items'] if playlist['name'] == playlist_name]
    if playlists:
        playlist_id = playlists[0]['id']
    else:
        logging.info("Creating playlist " + playlist_name)
        playlist_id = spotify.user_playlist_create(user=spotify.current_user()['id'], name=playlist_name)['id']

    logging.info("Adding " + str(len(tracks)) + " tracks to " + username + "'s playlist " + playlist_name + ": " + str(tracks))

    spotify.user_playlist_replace_tracks(user=spotify.current_user()['id'],
                                         playlist_id=playlist_id,
                                         tracks=[track.spotify_id for track in tracks])



from lastipy.spotify.parse_spotify_tracks import parse_tracks
import spotipy
import logging


#TODO test
def get_tracks_in_playlists(spotify, username):
    """Returns all tracks in the given user's playlists"""

    logging.info("Fetching all tracks in " + username + "'s playlists")

    playlists = spotify.current_user_playlists()['items']

    tracks = []
    for playlist in playlists:
        tracks = tracks + _get_tracks_in_playlist(spotify, username, playlist)

    logging.info("Fetched tracks " + str(tracks))

    return tracks


def _get_tracks_in_playlist(spotify, username, playlist):
    tracks_in_playlist = []

    keep_fetching = True
    while keep_fetching:
        json_response = spotify.user_playlist_tracks(user=username,
                                                   playlist_id=playlist['id'],
                                                   offset=len(tracks_in_playlist))
        if json_response['items']:
            tracks_in_playlist = tracks_in_playlist + parse_tracks(json_response['items'])
        else:
            keep_fetching = False

    return tracks_in_playlist


def get_saved_tracks(spotify, username):
    """Returns the all of the given user's saved tracks"""

    logging.info("Fetching " + username + "'s saved tracks")

    saved_tracks = []
    keep_fetching = True
    while keep_fetching:
        json_response = spotify.current_user_saved_tracks(offset=len(saved_tracks))
        if json_response['items']:
            saved_tracks = saved_tracks + parse_tracks(json_response['items'])
        else:
            keep_fetching = False

    logging.info("Fetched tracks " + str(saved_tracks))

    return saved_tracks



import spotipy
from lastipy.spotify.parse_spotify_tracks import parse_tracks
from datetime import datetime


def get_tracks_from_followed_artists(spotify, as_of_date):
    followed_artists = []

    curr_response = spotify.current_user_followed_artists(limit=50)

    while len(curr_response['artists']['items']) > 0:
        curr_response = spotify.current_user_followed_artists(limit=50, after=curr_response['artists']['items'][len(curr_response) - 1]['id'])
        followed_artists += curr_response['artists']['items']

    # The above Spotipy function doesn't really function properly and results in duplicates, 
    # so we remove them here by converting the list to just the IDs (not doing so results in
    # an unhashable error), then converting to a set and back to a list 
    followed_artists = [artist['id'] for artist in followed_artists]
    followed_artist_ids = list(set(followed_artists))

    all_albums = []
    for artist_id in followed_artist_ids:
        curr_response = spotify.artist_albums(artist_id, album_type='album', limit=50)
        artist_albums = curr_response['items']
        while len(curr_response['items']) > 0:
            curr_response = spotify.artist_albums(artist_id, album_type='album', limit=50, offset=len(artist_albums))
            artist_albums += curr_response['items']
        all_albums += artist_albums

    new_albums = []
    for album in all_albums:
        if album['release_date_precision'] == 'day':
            if datetime.strptime(album['release_date'], "%Y-%m-%d").date() >= as_of_date:
                new_albums.append(album)
        elif album['release_date_precision'] == 'month':
            release_date = datetime.strptime(album['release_date'], "%Y-%m")
            if release_date.year > as_of_date.year or (release_date.year == as_of_date.year and release_date.month >= as_of_date.month):
                   new_albums.append(album)
        elif album['release_date_precision'] == 'year':
            if datetime.strptime(album['release_date'], '%Y').year >= as_of_date.year:
                new_albums.append(album)

    all_tracks = []
    for album in new_albums:
        curr_response = spotify.album_tracks(album['id'], limit=50)
        album_tracks = curr_response['items']
        while len(curr_response['items']) > 0:
            curr_response = spotify.album_tracks(album['id'], limit=50, offset=len(album_tracks))
            album_tracks += curr_response['items']
        all_tracks += album_tracks

    return parse_tracks(all_tracks)



import logging
import spotipy
from lastipy.spotify import token
from lastipy.spotify.parse_spotify_tracks import parse_tracks
import json


#TODO test
def search_for_tracks(spotify, username, query):
    """Returns a list of tracks for the given query"""

    json_response = spotify.search(q=query)
    logging.debug("All search results for query " + query + ": " + json.dumps(json_response))

    converted_tracks = []
    if json_response['tracks']:
        converted_tracks = parse_tracks(json_response['tracks']['items'])
    logging.info("Track results for query " + query + ": " + str(converted_tracks))
    return converted_tracks
